#include "storedip"

struct ClientData {
    int targetbackend;
    time_t lastaccess;
};

struct ClientDataCmp {
    bool operator() (struct in_addr a, struct in_addr b) const {
	long la, lb;
	memcpy (&la, &a, sizeof(long));
	memcpy (&lb, &b, sizeof(long));
	return (la - lb) < 0;
    }
};

typedef  map<struct in_addr, ClientData, ClientDataCmp> StoreMap;
static StoreMap store;

unsigned StoredIp::target(struct in_addr clientip,
			  BackendVector const &targetlist) {
    unsigned target;
    time_t now = time(0);

    if (store.count(clientip) > 0) {
	// Client already known, maybe timed out.
	time_t diff = now - store[clientip].lastaccess;

	if (config.verbose()) {
	    Timestamp tm(store[clientip].lastaccess);
	    msg(Mstr("Client IP ") + Mstr(inet_ntoa(clientip)) +
		" last visited on " + tm.desc() +
		Mstr(Mstr(", ") + diff) + " sec ago, and went to " +
		balancer.backend(store[clientip].targetbackend).description() +
		"\n");
	}
	
	if (diff <= config.ipstoretimeout()) {
	    // Recent 'nuff
	    target = store[clientip].targetbackend;
	    if (! balancer.backend(target).available()) {
		// Historical target down - get new one if in lax mode
		if (config.dispatchmode() == Dispatchmode::m_strict_stored_ip)
		    throw static_cast<Error>
			("Stored-IP algorithm: target back end ") +
			balancer.backend(target).description() +
			"unavailable";
		else {
		    msg ("Stored IP algorithm: target back end " +
			 balancer.backend(target).description()  +
			 " unavailable, falling back to least-connections\n");
		    Leastconn l;
		    target = l.target(clientip, targetlist);
		}
	    }
	} else {
	    // Not recent anymore
	    msg ("Visit too long ago, re-dispatching with least-connections\n");
	    Leastconn l;
	    target = l.target(clientip, targetlist);
	}
    } else {
	// Historical target unknown, fetch new one
	msg ("New visit from " + static_cast<string>(inet_ntoa(clientip)) +
	     "\n");
	Leastconn l;
	target = l.target(clientip, targetlist);
    }

    // Update the info.
    ClientData entry = {target, now};
    store[clientip] = entry;

    // Weed out store.
    for (StoreMap::iterator iter = store.begin(); iter != store.end();
	 iter++) {
	if (config.debug()) {
	    Timestamp tm((*iter).second.lastaccess);
	    debugmsg (Mstr(inet_ntoa(iter->first)) + Mstr(" visited on ") +
		      tm.desc() + "\n");
	}
	if (now - ((*iter).second.lastaccess) > config.ipstoretimeout()) {
	    debugmsg ("  Erasing stale entry, stale\n");
	    store.erase(iter);
	}
    }

    // Return target to caller
    return (target);
}
