#include "balancer"
#include "../tcpdispatcher/tcpdispatcher"
#include "../httpdispatcher/httpdispatcher"

void Balancer::serve() {
    int clsock;

    // Start up wakeup/checkup handlers. These are always started - even
    // when config.wakeupsec() and config.checkupsec() are not defined
    // and have value 0. Via the web interface, the values can be later
    // changed, but we want to have the checkers running always.
    if (!config.foregroundmode() && config.sport()) {
	msg ("Starting wakeup thread.\n");
	Wakeupthread *wt = new Wakeupthread();
	if (!wt)
	    throw static_cast<Error>("Memory fault in Balancer::serve");
	wt->start();
	
	msg ("Starting checkup thread.\n");
	Checkupthread *ct = new Checkupthread();
	if (!ct)
	    throw static_cast<Error>("Memory fault in Balancer::serve");
	ct->start();
    }

    // Write the PID file.
    if (config.pidfile() != "") {
	FILE *f;
	if (! (f = fopen (config.pidfile().c_str(), "w")) )
	    throw static_cast<Error>("Cannot write pid file ") +
		config.pidfile() + ": " + strerror(errno);
	fprintf (f, "%u\n", getpid());
	fclose (f);
    }

    // Wait for activity, serve it.
    msg ((Mstr("Awaiting activity on fd ") + server_fd) + "\n");
    while (true) {
	Fdset fdset(0);
	fdset.add (server_fd);
	if (fdset.readable() < 0) {
	    // We caught a signal. Either a request to report status,
	    // or to terminate.
	    msg ("Interrupt seen\n");
	    if (terminate()) {
		msg ("Termination requested, XR will stop.\n");
		break;
	    } else if (report()) {
		msg ("Report requested\n");
		reportmsg ("*** XR STATUS REPORT STARTS ***\n");
		for (unsigned i = 0; i < nbackends(); i++) {
		    reportmsg("Back end " + backend(i).description() + "\n");
		    reportmsg((Mstr("  Status: ") +
			       backend(i).availablestr()) +
			      (Mstr(", ") + backend(i).livestr()) + "\n");
		    reportmsg((Mstr("  Connections: ") +
			       backend(i).connections())
			       +
			       (Mstr(" (max ") + backend(i).maxconn()) + "\n");
		    reportmsg((Mstr("  Served:") + backend(i).bytesserved()) +
			      (Mstr(" bytes, ") + backend(i).clientsserved()) +
			      " clients\n");
		}
		report (false);
		reportmsg ("*** XR STATUS REPORT ENDS ***\n");
		continue;
	    } else {
		msg ("Non-meaningful interrupt, resuming\n");
		continue;
	    }
	}

	// Got activity!
	request_nr++;

	if (server_fd) {
	    // If tcp-serving: server_fd > 0; serve and loop again
	    int size;
	    struct sockaddr_in clname;

	    size = sizeof(clname);
	    if ( (clsock = accept (server_fd, (struct sockaddr *) &clname,
				   (socklen_t*) &size)) < 0 ) {
		warnmsg(Mstr("Failed to accept network connection: ") +
			Mstr(strerror(errno)) + "\n");
		continue;
	    }

	    string clientip = inet_ntoa(clname.sin_addr);

	    // If there is an allow list, the client must match it.
	    if (config.nallow()) {
		if (config.debug())
		    debugmsg ("Matching " + clientip + " against allow list\n");
		bool allowed = false;
		for (unsigned n = 0; n < config.nallow(); n++) {
		    if (ipmatch (clname.sin_addr, config.allow(n))) {
			allowed = true;
			break;
		    }
		}
		if (!allowed) {
		    msg ("Not serving client " + clientip +
			 ": no match in allow list\n");
		    socketclose (clsock);
		    continue;
		}
	    }
	    // If the client is in the deny list, deny it.
	    if (config.debug())
		debugmsg ("Matching " + clientip + " against deny list\n");
	    bool denied = false;
	    for (unsigned n = 0; n < config.ndeny(); n++)
		if (ipmatch (clname.sin_addr, config.deny(n))) {
		    denied = true;
		    break;
		}
	    if (denied) {
		msg ("Not serving client " + clientip +
		     ": match in deny list\n");
		socketclose (clsock);
		continue;
	    }

	    // Show whom we've accepted
	    if (config.verbose()) {
		msg ((Mstr("Accepted connection from ") + clientip) +
		     (Mstr(" as client fd ") + clsock) + "\n");
		msg ((Mstr("Balancer is serving ") + connections()) +
		     " clients\n");
		msg ("Current back end states:\n");
		for (unsigned i = 0; i < nbackends(); i++)
		    msg((Mstr("  Back end ") + backend(i).description()) +
			(Mstr(": ") + backend(i).connections()) +
			(Mstr(" connections , max ") + backend(i).maxconn()) +
			(Mstr(", status ") + backend(i).availablestr()) + "\n");
	    }

	    // We got action! Check if the total connections to the
	    // balancer doesn't exceed the max.
	    if (config.maxconn() && connections() >= config.maxconn()) {
		msg ((Mstr("Not serving client: already ") + connections()) +
		     (Mstr(" (max ") + config.maxconn()) + "\n");
		socketclose (clsock);
		continue;
	    }

	    TcpDispatcher *d;
	    switch (config.stype()) {
	    case Servertype::t_tcp:
		d = new TcpDispatcher (clsock, clname.sin_addr);
		break;
	    case Servertype::t_http:
		d = new HttpDispatcher (clsock, clname.sin_addr);
		break;
	    default:
		throw static_cast<Error>("Internal error, "
					 "can't choose dispatcher");
		break;
	    }
	    if (!d)
		throw static_cast<Error>("Memory fault in Balancer::serve");

	    // Allocation boundary printout
	    if (config.debug()) {
		void *mem = malloc(16);
		free (mem);
		_debugmsg (Mstr("Allocation boundary at dispatcher start: ") +
			  mem + "\n");
	    }

	    d->start();
	} else {
	    // If fd-serving, serve and close. Don't thread it up.
	    TcpDispatcher *d;
	    struct in_addr dummy;
	    inet_aton ("0.0.0.0", &dummy);

	    switch (config.stype()) {
	    case Servertype::t_tcp:
		d = new TcpDispatcher (server_fd, dummy);
		break;
	    case Servertype::t_http:
		d = new HttpDispatcher (server_fd, dummy);
		break;
	    default:
		throw static_cast<Error>("Internal error, "
					 "can't choose dispatcher");
		break;
	    }
	    if (!d)
		throw static_cast<Error>("Memory fault in Balancer::serve");
	    d->execute();
	    break;
	}

	// If we exceed the max # of requests, stop..
	if (config.quitafter()) {
	    msg ((Mstr("Request ") + requestnr()) +
		 (Mstr(" underway of max ") + config.quitafter()) + "\n");
	    if (requestnr() >=  (long)config.quitafter()) {
		msg ("Max requests served, will stop.\n");
		break;
	    }
	}
    }

    // We're stopping now. If a PID stamp was created, remove it.
    if (config.pidfile() != "")
	unlink (config.pidfile().c_str());

    // Wait for running threads to die off.
    socketclose (server_fd);
    shutdown (server_fd, SHUT_RDWR);
    unsigned prev_conn = 0x19081962;
    while (1) {
	unsigned curr_conn = balancer.connections();
	if (!curr_conn)
	    break;
	if (curr_conn != prev_conn) {
	    msg ((Mstr("There are still ") + curr_conn) + " connections\n");
	    prev_conn = curr_conn;
	}
	sleep (1);
    }
    msg ("XR is idle, stopping.\n");
}
