#include "httpdispatcher"

void HttpDispatcher::handle() {
    PROFILE("HttpDispatcher::handle");

    // The client request was already retrieved before starting the
    // dispatcher. We can continue by applying server-directed headers.
    if (config.addxrversion())
	buf.setheader ("XR", VER);
    if (config.addxforwardedfor())
	buf.addheader ("X-Forwarded-For", string(inet_ntoa(clientip())));
    for (unsigned n = 0; n < config.nserverheaders(); n++)
	buf.setheader (config.serverheader(n));

    // Flush client info received so far to the back end.
    debugmsg("Sending client request to back end\n");
    buf.netwrite(backendfd(), config.backend_timeout());

    // Let's see if we will need to modify the server headers.
    bool modify_serverheaders = false;
    if (config.addxrversion() ||
	(config.stickyhttp() && !issticky()))
	modify_serverheaders = true;

    // Store the client request. May want to log it later.
    string client_request = buf.firstline();

    // Go into copy-thru mode. If required, catch the server headers on
    // their first appearance and modify them.
    bool backend_response_checked = false;
    while (1) {
	Fdset readset (config.client_timeout());
	readset.add(clientfd());
	readset.add(backendfd());

	int sock;
	if ((sock = readset.readable()) < 0)
	    break;
	
	buf.reset();
	
	if (!buf.netread(sock))
	    break;

	if (sock == backendfd() && modify_serverheaders) {
	    debugmsg("Back end response seen, applying modifications\n");
	    modify_serverheaders = false;
	    while (! buf.headersreceived())
		if (!buf.netread (sock, config.backend_timeout()))
		    throw static_cast<Error>
			("Failed to get headers from back end");
	    if (config.addxrversion())
		buf.setheader("XR", VER);
	    if (config.stickyhttp() && !issticky()) {
		ostringstream o;
		o << "XRTarget=" << targetbackend();
		buf.setheader("Set-Cookie", o.str());
	    }
	}
	
	// The back end response may now get flushed to the client. 
	// If the response code is 4** or 5**, log it as a warning.
	if (!backend_response_checked &&
	    sock == backendfd() && buf.headersreceived()) {
	    string respcode = buf.stringat(9, 3);
	    if (respcode[0] == '4' || respcode[0] == '5')
		warnmsg("HTTP back end indicates fault: '" +
			buf.firstline() + "' as response to '" +
			client_request + "'\n");
	    backend_response_checked = true;
	}

	// Flush info to the other connected side.
	int othersock, timeout;
	if (sock == clientfd()) {
	    othersock = backendfd();
	    timeout = config.backend_timeout();
	} else {
	    othersock = clientfd();
	    timeout = config.client_timeout();
	}

	debugmsg (Mstr("Had data on ") + sock +
		  (Mstr(", sending to ") + othersock) + "\n");

	buf.netwrite(othersock, timeout);
	if (sock == backendfd())
	    balancer.backend(targetbackend()).addbytes(buf.bufsz());
    }

}
