
#include "sys/sys"
#include "config/config"
#include "balancer/balancer"
#include "error/error"
#include "profiler/profiler"

using namespace std;

Config config;
Balancer balancer;

static void showlimits() {
    typedef struct {
	int resource;
	string description;
    } Limit;
    static Limit limit[] = {
	{ RLIMIT_CORE, 		"coredump size (bytes)" },
	{ RLIMIT_CPU,		"cpu time (sec)" },
	{ RLIMIT_DATA,		"data segment size (bytes)" },
	{ RLIMIT_FSIZE,		"max file size (bytes)" },
#	ifdef RLIMIT_MEMLOCK	
	{ RLIMIT_MEMLOCK,	"locked mem size (bytes)" },
#	endif	
	{ RLIMIT_NOFILE,	"max open files" },
#	ifdef RLIMIT_NPROC	
	{ RLIMIT_NPROC,		"max processes" },
#	endif
#	ifdef RLIMIT_RSS	
	{ RLIMIT_RSS,		"max resident set size (bytes)" },
#	endif	
	{ RLIMIT_STACK,		"max stack size (bytes)" }
    };
    
    for (unsigned i = 0; i < sizeof(limit) / sizeof(Limit); i++) {
	struct rlimit rl;
	if (getrlimit(limit[i].resource, &rl))
	    throw static_cast<Error>("Failed to request limit: ") +
		strerror(errno);
	ostringstream o;
	o << "Limits for " << limit[i].description
	  << ": hard limit " << unsigned(rl.rlim_max)
	  << ", soft limit " << unsigned(rl.rlim_cur) << '\n';
	msg(o.str());
    }
}

static void sigcatcher (int sig) {
    debugmsg ("Seen signal " + sig + '\n');
    if (sig == SIGHUP)
	balancer.report(true);
    else if (sig != SIGPIPE && sig != SIGSTOP)
	balancer.terminate(true);
    // Actually we wouldn't need to test for SIGPIPE, it's ignored (see below).
    // Leaving the test in place for future versions, better an extra if
    // than forgetting it later.
    // SIGSTOP is used for stopping separare threads.
}

int main (int argc, char **argv) {

    PROFILE("main");
    
    static int relevant_sig[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGABRT, SIGTERM, SIGSTOP
    };
    
    try {
	// Load configuration from the commandline, promote verbosity
	config.parsecmdline (argc, argv);

	if (config.verbose())
	    showlimits();
	
	msg ((Mstr("XR running as PID ") + getpid()) + "\n");
	
	// Load the signal handler.
	for (unsigned i = 0; i < sizeof(relevant_sig) / sizeof(int); i++)
	    signal (relevant_sig[i], sigcatcher);
	signal (SIGPIPE, SIG_IGN);
	
	// Configure the balancer and start serving.
	balancer.init();
	balancer.serve();
    } catch (Error const &e) {
	cerr << e.what() << endl;
	return (1);
    } catch (...) {
	cerr << "ERROR: Unidentified exception caught" << endl;
    }
    
    return (0);
}
